﻿using System;
using System.Collections.Generic;
using DES.PopulationGrowth.Events;
using DES.PopulationGrowth.Objects;
using MathNet.Numerics.Distributions;
using MathNet.Numerics.Random;
using Puzzle.MicroHeap;

namespace DES.PopulationGrowth
{
    public class Simulation
    {
        public List<Individual> Population { get; set; }
        public int Time { get; set; }
        private int _currentTime;
        private readonly Dictionary<Event, IDistribution> _distributions ;

        public Simulation(IEnumerable<Individual> population, int time)
        {
            Population = new List<Individual>(population);
            Time = time;
            _distributions = new Dictionary<Event, IDistribution>
                                 {
                                     { Event.CapableEngaging, new Poisson(18) },
                                     { Event.BirthEngageDisengage, new ContinuousUniform() },
                                     { Event.GetPregnant, new Normal(28, 8) },
                                     { Event.ChildrenCount, new Normal(2, 6) },
                                     { Event.TimeChildren, new Exponential(8) },
                                     { Event.Die, new Poisson(70) },
                                 };
            
            foreach (var individual in Population)
            {
                // LifeTime
                individual.LifeTime = ((Poisson) _distributions[Event.Die]).Sample();
                // Ready to start having relations
                individual.RelationAge = ((Poisson)_distributions[Event.CapableEngaging]).Sample();
                // Pregnancy Age (only women)
                if (individual is Female)
                {
                    (individual as Female).PregnancyAge = ((Normal) _distributions[Event.GetPregnant]).Sample();
                    (individual as Female).ChildrenCount = ((Normal)_distributions[Event.ChildrenCount]).Sample();
                }
            }
        }

        public void Execute()
        {
            while (_currentTime < Time)
            {
                // Check what happens to every individual this year
                for (var i = 0; i < Population.Count; i++)
                {
                    var individual = Population[i];

                    // Event -> Birth
                    if (individual is Female && (individual as Female).IsPregnant)
                        Population.Add((individual as Female).GiveBirth(_distributions, _currentTime));

                    // Event -> Check whether someone starts a relation this year
                    if (individual.SuitableRelation())
                        individual.FindCouple(Population, _currentTime, _distributions);

                    // Events where having an engaged individual represents a prerequisite
                    if (individual.Engaged)
                    {
                        // Event -> Check whether some relation ends this year
                        if (individual.EndRelation(_distributions))
                            individual.Disengage();

                        // Event -> Check whether some couple can have a child now
                        if (individual is Female &&
                        (individual as Female).SuitablePregnancy(_currentTime))
                            (individual as Female).IsPregnant = true;
                    }

                    // Event -> Check whether someone dies this year
                    if (individual.Age.Equals(individual.LifeTime))
                    {
                        // Case: Individual in relation (break relation)
                        if (individual.Engaged)
                            individual.Disengage();
                        Population.RemoveAt(i);
                    }

                    individual.Age++;
                    _currentTime++;
                }
            }
        }
    }
}